from pathlib import Path

from yarl import URL

# local path where the package stores its data
CACHE_PATH = Path.home() / ".genomic_benchmarks"
# local path where the references are stored
REF_CACHE_PATH = CACHE_PATH / "fasta"
# where the interval datasets are stored (if read from local)
DATASET_DIR_PATH = (Path(__file__).parents[0] / ".." / ".." / ".." / "datasets").resolve()
# where the interval datasets are stored (if read from the internet)
DATASET_URL_PATH = URL("http://raw.githubusercontent.com/ML-Bioinfo-CEITEC/genomic_benchmarks/main/datasets/")



import os

import numpy as np

import torch
from torch import nn


# A simple CNN model
class CNN(nn.Module):
    def __init__(self, number_of_classes, vocab_size, embedding_dim, input_len):
        super(CNN, self).__init__()
        if number_of_classes == 2:
            number_of_output_neurons = 1
            loss = torch.nn.functional.binary_cross_entropy_with_logits
            output_activation = nn.Sigmoid()
        else:
            raise Exception("Not implemented for number_of_classes!=2")
            # number_of_output_neurons = number_of_classes
            # loss = torch.nn.CrossEntropyLoss()
            # output_activation = nn.Softmax(dim=)

        self.embeddings = nn.Embedding(vocab_size, embedding_dim)
        self.conv1 = nn.Conv1d(in_channels=embedding_dim, out_channels=16, kernel_size=8, bias=True)
        self.norm1 = nn.BatchNorm1d(16)
        self.relu = nn.ReLU()
        self.pool1 = nn.MaxPool1d(2)

        self.conv2 = nn.Conv1d(in_channels=16, out_channels=8, kernel_size=8, bias=True)
        self.norm2 = nn.BatchNorm1d(8)
        self.pool2 = nn.MaxPool1d(2)

        self.conv3 = nn.Conv1d(in_channels=8, out_channels=4, kernel_size=8, bias=True)
        self.norm3 = nn.BatchNorm1d(4)
        self.pool3 = nn.MaxPool1d(2)

        #         compute output shape of conv layers
        self.flatten = nn.Flatten()
        self.lin1 = nn.Linear(self.count_flatten_size(input_len), 512)
        self.lin2 = nn.Linear(512, number_of_output_neurons)
        self.output_activation = output_activation
        self.loss = loss

    def count_flatten_size(self, input_len):
        zeros = torch.zeros([1, input_len], dtype=torch.long)
        x = self.embeddings(zeros)
        x = x.transpose(1, 2)
        x = self.conv1(x)
        x = self.norm1(x)
        x = self.relu(x)
        x = self.pool1(x)

        x = self.conv2(x)
        x = self.norm2(x)
        x = self.relu(x)
        x = self.pool2(x)

        x = self.conv3(x)
        x = self.norm3(x)
        x = self.relu(x)
        x = self.pool3(x)

        x = self.flatten(x)
        return x.size()[1]

    def forward(self, x):
        x = self.embeddings(x)
        x = x.transpose(1, 2)
        x = self.conv1(x)
        x = self.norm1(x)
        x = self.relu(x)
        x = self.pool1(x)

        x = self.conv2(x)
        x = self.norm2(x)
        x = self.relu(x)
        x = self.pool2(x)

        x = self.conv3(x)
        x = self.norm3(x)
        x = self.relu(x)
        x = self.pool3(x)

        x = self.flatten(x)
        x = self.lin1(x)
        x = self.lin2(x)
        x = self.output_activation(x)
        return x

    def train_loop(self, dataloader, optimizer):
        for x, y in dataloader:
            optimizer.zero_grad()
            pred = self(x)
            loss = self.loss(pred, y)
            loss.backward()
            optimizer.step()

        #       train acc
        # todo: optimize counting of acc
        size = dataloader.dataset.__len__()
        num_batches = len(dataloader)
        train_loss, correct = 0, 0

        with torch.no_grad():
            for X, y in dataloader:
                pred = self(X)
                train_loss += self.loss(pred, y).item()
                correct += (torch.round(pred) == y).sum().item()

        train_loss /= num_batches
        correct /= size
        print(f"Train metrics: \n Accuracy: {(100*correct):>0.1f}%, Avg loss: {train_loss:>8f} \n")

    def train(self, dataloader, epochs):
        optimizer = torch.optim.Adam(self.parameters())
        for t in range(epochs):
            print(f"Epoch {t}")
            self.train_loop(dataloader, optimizer)

    def test(self, dataloader):
        size = dataloader.dataset.__len__()
        num_batches = len(dataloader)
        test_loss, correct = 0, 0

        with torch.no_grad():
            for X, y in dataloader:
                pred = self(X)
                test_loss += self.loss(pred, y).item()
                correct += (torch.round(pred) == y).sum().item()

        print("test_loss ", test_loss)
        print("num_batches", num_batches)
        print("correct", correct)
        print("size", size)

        test_loss /= num_batches
        correct /= size
        print(f"Test Error: \n Accuracy: {(100*correct):>0.1f}%, Avg loss: {test_loss:>8f} \n")



import re
import shutil  # for removing and creating folders
from pathlib import Path

import pandas as pd
from genomic_benchmarks.utils.datasets import (
    _check_dataset_existence,
    _download_url,
    _fastagz2dict,
    _get_dataset_name,
    _get_reference_name,
    _guess_location,
    _remove_and_create,
    _rev,
)
from genomic_benchmarks.utils.paths import CACHE_PATH, REF_CACHE_PATH

from .cloud_caching import CLOUD_CACHE, download_from_cloud_cache


def download_dataset(
    interval_list_dataset,
    version=None,
    dest_path=CACHE_PATH,
    cache_path=REF_CACHE_PATH,
    force_download=False,
    use_cloud_cache=True,
    local_repo=False,
):
    """
    Transform an interval-list genomic dataset into a full-seq genomic dataset.

            Parameters:
                    interval_list_dataset (str or Path): Either a path or a name of dataset included in this package.
                    version (int): Version of the dataset.
                    dest_path (str or Path): Folder to store the full-seq dataset.
                    cache_path (str or Path): Folder to store the downloaded references.
                    force_download (bool): If True, force downloading of references.
                    use_cloud_cache (bool): If True, use the cloud cache for downloading a full-seq genomic datasets.
                    local_repo (bool): If True, use the local repo for getting interval-list genomic datasets.

            Returns:
                    seq_dataset_path (Path): Path to the full-seq dataset.
    """

    interval_list_dataset = _guess_location(interval_list_dataset, local_repo)
    metadata = _check_dataset_existence(interval_list_dataset, version, local_repo)
    dataset_name = _get_dataset_name(interval_list_dataset)

    if version is None:
        version = metadata["version"]
    if use_cloud_cache and ((dataset_name, version) in CLOUD_CACHE):
        Path(dest_path).mkdir(parents=True, exist_ok=True)  # to be sure "./.genomic_benchmarks" exists
        return download_from_cloud_cache((dataset_name, version), Path(dest_path) / dataset_name)

    refs = _download_references(metadata, cache_path=cache_path, force=force_download)
    fastas = _load_fastas_into_memory(refs, cache_path=cache_path)

    _remove_and_create(Path(dest_path) / dataset_name)
    _remove_and_create(Path(dest_path) / dataset_name / "train")
    _remove_and_create(Path(dest_path) / dataset_name / "test")

    for c in metadata["classes"]:
        for t in ["train", "test"]:
            dt_filename = interval_list_dataset / t / (c + ".csv.gz")
            dt = pd.read_csv(str(dt_filename), compression="gzip")

            ref_name = _get_reference_name(metadata["classes"][c]["url"])
            dt["seq"] = _fill_seq_column(fastas[ref_name], dt)

            folder_filename = Path(dest_path) / dataset_name / t / c
            _remove_and_create(folder_filename)
            for row in dt.iterrows():
                row_filename = folder_filename / (str(row[1]["id"]) + ".txt")
                row_filename.write_text(row[1]["seq"])

    return Path(dest_path) / dataset_name


EXTRA_PREPROCESSING = {
    # known extra preprocessing steps
    "default": [None, None, lambda x: x],
    "ENSEMBL_HUMAN_GENOME": [24, "MT", lambda x: "chr" + x],  # use only chromosomes, not contigs, and add chr prefix
    "ENSEMBL_MOUSE_GENOME": [21, "MT", lambda x: "chr" + x],  # use only chromosomes, not contigs, and add chr prefix
    "ENSEMBL_HUMAN_TRANSCRIPTOME": [
        190_000,
        None,
        lambda x: re.sub("ENST([0-9]*)[.][0-9]*", "ENST\\1", x),
    ],  # remove the version number from the ensembl id
}


def _load_fastas_into_memory(refs, cache_path):
    # load all references into memory
    fastas = {}
    for ref in refs:
        ref_path = Path(cache_path) / _get_reference_name(ref[0])
        ref_type = ref[1]
        ref_extra_preprocessing = ref[2] if ref[2] is not None else "default"
        if ref_extra_preprocessing not in EXTRA_PREPROCESSING:
            raise ValueError(f"Unknown extra preprocessing: {ref_extra_preprocessing}")

        if ref_type == "fa.gz":
            fasta = _fastagz2dict(
                ref_path,
                fasta_total=EXTRA_PREPROCESSING[ref_extra_preprocessing][0],
                stop_id=EXTRA_PREPROCESSING[ref_extra_preprocessing][1],
                region_name_transform=EXTRA_PREPROCESSING[ref_extra_preprocessing][2],
            )
            fastas[_get_reference_name(ref[0])] = fasta
        else:
            raise ValueError(f"Unknown reference type {ref_type}")
    return fastas


def _download_references(metadata, cache_path, force=False):
    # download all references from the metadata into cache_path folder
    cache_path = Path(cache_path)
    if not cache_path.exists():
        cache_path.mkdir(parents=True)

    refs = {(c["url"], c["type"], c.get("extra_processing")) for c in metadata["classes"].values()}

    for ref in refs:
        ref_path = cache_path / _get_reference_name(ref[0])
        if not ref_path.exists() or force:
            _download_url(ref[0], ref_path)
        else:
            print(f"Reference {ref_path} already exists. Skipping.")

    return refs


def _fill_seq_column(fasta, df):
    # fill seq column in DataFrame tab
    if not all([r in fasta for r in df["region"]]):
        missing_regions = list({r for r in df["region"] if r not in fasta})
        if len(missing_regions) > 5:
            missing_regions = missing_regions[:6]
        raise ValueError("Some regions not found in the reference, e.g. " + " ".join([str(r) for r in missing_regions]))
    output = pd.Series(
        [
            _rev(fasta[region][start:end], strand)
            for region, start, end, strand in zip(df["region"], df["start"], df["end"], df["strand"])
        ]
    )
    return output


def remove_dataset_from_disk(interval_list_dataset, version=None, dest_path=CACHE_PATH):
    """
    Remove the full-seq dataset from the disk.

            Parameters:
                    interval_list_dataset (str or Path): Either a path or a name of dataset included in this package.
                    version (int): Version of the dataset.
                    dest_path (str or Path): Folder to store the full-seq dataset.
    """
    interval_list_dataset = _guess_location(interval_list_dataset)
    metadata = _check_dataset_existence(interval_list_dataset, version)
    dataset_name = _get_dataset_name(interval_list_dataset)

    path = Path(dest_path) / dataset_name
    if path.exists():
        shutil.rmtree(path)



import tensorflow as tf
from tensorflow.keras.layers import (
    BatchNormalization,
    Conv1D,
    Dense,
    Dropout,
    GlobalAveragePooling1D,
    MaxPooling1D,
)
from tensorflow.keras.layers.experimental.preprocessing import TextVectorization

character_split_fn = lambda x: tf.strings.unicode_split(x, "UTF-8")
vectorize_layer = TextVectorization(output_mode="int", split=character_split_fn)

# one-hot encoding
onehot_layer = tf.keras.layers.Lambda(lambda x: tf.one_hot(tf.cast(x, "int64"), 4))

basic_cnn_model_v0 = tf.keras.Sequential(
    [
        onehot_layer,
        Conv1D(32, kernel_size=8, data_format="channels_last", activation="relu"),
        BatchNormalization(),
        MaxPooling1D(),
        Conv1D(16, kernel_size=8, data_format="channels_last", activation="relu"),
        BatchNormalization(),
        MaxPooling1D(),
        Conv1D(4, kernel_size=8, data_format="channels_last", activation="relu"),
        BatchNormalization(),
        MaxPooling1D(),
        Dropout(0.3),
        GlobalAveragePooling1D(),
        Dense(1),
    ]
)



import shutil
from pathlib import Path

from google_drive_downloader import (
    GoogleDriveDownloader as gdd,  # TODO: use gdown instead of google_drive_downloader
)

CLOUD_CACHE = {
    ("human_nontata_promoters", 0): "1VdUg0Zu8yfLS6QesBXwGz1PIQrTW3Ze4",
    ("demo_coding_vs_intergenomic_seqs", 0): "1cpXg0ULuTGF7h1_HTYvc6p8M-ee43t-v",
    ("demo_human_or_worm", 0): "1Vuc44bXRISqRDXNrxt5lGYLpLsJbrSg8",
    ("demo_mouse_enhancers", 0): "1u3pyaL8smQaJXeOx7YZkjj-Bdpb-jGCM",
    ("human_enhancers_cohn", 0): "176563cDPQ5Y094WyoSBF02QjoVQhWuCh",
    ("human_enhancers_ensembl", 0): "1gZBEV_RGxJE8EON5OObdrp5Tp8JL0Fxb",
}


def download_from_cloud_cache(file_key, dest_path, cloud_cache=CLOUD_CACHE, force_download=True):
    """
    Download a file from the cloud cache.
    """
    if file_key not in cloud_cache:
        raise ValueError(f"File ID {file_key} not in the cloud cache.")

    if force_download and Path(dest_path).exists():
        shutil.rmtree(str(dest_path))

    gdd.download_file_from_google_drive(file_id=cloud_cache[file_key], dest_path=str(dest_path) + ".zip", unzip=True)
    Path(str(dest_path) + ".zip").unlink()

    return Path(dest_path)



import gzip
import shutil  # for removing and creating folders
import urllib
import warnings
from pathlib import Path

import requests
import yaml
from Bio import SeqIO
from Bio.Seq import Seq
from tqdm.autonotebook import tqdm
from yarl import URL

from .paths import DATASET_DIR_PATH, DATASET_URL_PATH


def _guess_location(dataset_path, local_repo: bool = False):
    if Path(dataset_path).exists():
        return Path(dataset_path)
    elif local_repo and (DATASET_DIR_PATH / str(dataset_path)).exists():
        return DATASET_DIR_PATH / str(dataset_path)
    elif (not local_repo) and requests.get(DATASET_URL_PATH / str(dataset_path) / "metadata.yaml").status_code == 200:
        return DATASET_URL_PATH / str(dataset_path)
    else:
        raise FileNotFoundError(f"Dataset {dataset_path} not found.")


def _check_dataset_existence(interval_list_dataset, version, local_repo: bool = False):
    # check that the dataset exists, returns its metadata

    if local_repo:
        path = Path(interval_list_dataset)
        if not path.exists():
            raise FileNotFoundError(f"Dataset {interval_list_dataset} not found.")

        metadata_path = path / "metadata.yaml"
        if not metadata_path.exists():
            raise FileNotFoundError(f"Dataset {interval_list_dataset} does not contain `metadata.yaml` file.")
        with open(metadata_path, "r") as fr:
            metadata = yaml.safe_load(fr)
    else:
        url = URL(interval_list_dataset)
        with urllib.request.urlopen(str(url / "metadata.yaml")) as fr:
            metadata = yaml.safe_load(fr)

    if version is not None:
        if version != metadata["version"]:
            raise ValueError(f"Dataset version {version} does not match the version in metadata {metadata['version']}.")
    else:
        warnings.warn(f"No version specified. Using version {metadata['version']}.")

    return metadata


def _get_dataset_name(path):
    # get the dataset name from the path
    return Path(str(path)).stem


def _get_reference_name(url):
    # get the reference name from the url
    # TODO: better naming scheme (e.g. taking the same file from 2 Ensembl releases)
    return url.split("/")[-1]


def _download_url(url, dest):
    # download a file from url to dest
    if Path(dest).exists():
        Path(dest).unlink()

    print(f"Downloading {url}")

    class DownloadProgressBar(tqdm):
        # for progress bar
        def update_to(self, b=1, bsize=1, tsize=None):
            if tsize is not None:
                self.total = tsize
            self.update(b * bsize - self.n)

    with DownloadProgressBar(unit="B", unit_scale=True, miniters=1, desc=str(dest)) as t:
        # TODO: adapt fastdownload code instead of urllib
        urllib.request.urlretrieve(url, filename=dest, reporthook=t.update_to)


def _fastagz2dict(fasta_path, fasta_total=None, stop_id=None, region_name_transform=lambda x: x):
    # load gzipped fasta into dictionary
    fasta = {}

    with gzip.open(fasta_path, "rt") as handle:
        for record in tqdm(SeqIO.parse(handle, "fasta"), total=fasta_total):
            fasta[region_name_transform(record.id)] = str(record.seq)
            if stop_id and (record.id == stop_id):
                # stop, do not read small contigs
                break
    return fasta


def _rev(seq, strand):
    # reverse complement
    if strand == "-":
        return str(Seq(seq).reverse_complement())
    else:
        return seq


def _remove_and_create(path):
    # cleaning step: remove the folder and then create it again
    if path.exists():
        shutil.rmtree(path)
    path.mkdir(parents=True)



from setuptools import find_packages, setup

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name="genomic_benchmarks",
    version="0.0.3",
    description="Genomic Benchmarks",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="RBP Bioinformatics",
    author_email="ML.Bioinfo.CEITEC@gmail.com",
    license="Apache License 2.0",
    keywords=["bioinformatics", "genomics", "data"],
    url="https://github.com/ML-Bioinfo-CEITEC/genomic_benchmarks",
    packages=find_packages("src"),
    package_dir={"": "src"},
    scripts=[],
    # setup_requires=['pytest-runner'], setup_requires deprecated in v58.3.0
    install_requires=[
        "biopython>=1.79",
        "requests>=2.23.0",
        "pip>=20.0.1",
        "numpy>=1.17.0",
        "pandas>=1.1.4",
        "tqdm>=4.41.1",
        "pyyaml>=5.3.1",
        "googledrivedownloader>=0.4",  # 'gdown>=4.2.0',
        "yarl",
        #'papermill>=2.3.0',
        #'tensorflow>=2.6.0',
        #'jupyter>=1.0.0',
        #'torch>=1.9.0',
    ],
    tests_require=["pytest"],
    include_package_data=True,
    package_data={},
    classifiers=[
        # Chose either "3 - Alpha", "4 - Beta" or "5 - Production/Stable" as the current state of your package
        "Development Status :: 3 - Alpha",
        # Define that your audience are developers
        "Intended Audience :: Developers",
        "Topic :: Software Development :: Build Tools",
        "License :: OSI Approved :: MIT License",
        # Specify which pyhton versions that you want to support
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
    ],
)
